// AppController.h -- implementation for the UndoQueue controller class
//                    (actually, the only custom class here)

#import "AppController.h"


@implementation AppController


// our nib file has been loaded and the connections have been
// hooked up.  Here we do our initialization stuff

- (void) awakeFromNib
{
    // initialize our 'model' based on what was set in the nib file
    thing = [slider floatValue];
    color = [[colorWell color] copy];

    // update the 'view'
    [displayField setTextColor: color];
    [displayField setIntValue: thing];

    // just for fun, register with all the undo notifications so we can
    // see what's happening under the hood
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];

    NSArray *notifications;
    notifications = [NSArray arrayWithObjects:
			     NSUndoManagerCheckpointNotification,
			     NSUndoManagerDidOpenUndoGroupNotification,
			     NSUndoManagerDidRedoChangeNotification,
			     NSUndoManagerDidUndoChangeNotification,
			     NSUndoManagerWillCloseUndoGroupNotification,
			     NSUndoManagerWillRedoChangeNotification,
			     NSUndoManagerWillUndoChangeNotification,
			     @"ThingChanged", // also show our own undo-related
			     @"ColorChanged", // notifications
			     nil];
    
    NSEnumerator *enumerator;
    enumerator = [notifications objectEnumerator];

    NSString *notification;
    while ((notification = [enumerator nextObject])) {
	[center addObserver: self
		selector: @selector(logNotification:)
		name: notification
		object: nil];
    }


    // now make the notification queue for the coalesced undo.
    // whenever a control is twidded (and coalesceP is set to YES),
    // rather than making an undo operation, instead post a
    // notification to this queue.  At idle time, all of the notifications
    // in this queue will get reduced to a single notification that
    // is posted to the default notification center
    undoQueue = [[NSNotificationQueue alloc]
		    initWithNotificationCenter: center];

    // these are the notifications that will come from the undoQueue 
    [center addObserver: self
	    selector: @selector(thingChanged:)
	    name: @"ThingChanged"
	    object: nil];

    [center addObserver: self
	    selector: @selector(colorChanged:)
	    name: @"ColorChanged"
	    object: nil];

} // awakeFromNib



// clean up our mess

- (void) dealloc
{
    [color release];
    [undoQueue release];

    // remove ourselves as observers in the default notification center
    NSNotificationCenter *center;
    center = [NSNotificationCenter defaultCenter];

    [center removeObserver: self];

    // make sure our parent class cleans up their mess too
    [super dealloc];

} // dealloc



// --------------------------------------------------

// the next three functions are the undo handlers for the slider
// which changes the 'thing' value.
//  * changeThing:  is what takes a new float value.  It will change the
//                  model, and make a new undo action

//  * thingChanged: this is the notification method that the default
//                  notification center will call after our undo queue has
//                  posted the notification (hmmm.. that's clear as mud)
//                  This calls changeThing: to actually change the value
//                  and create an undo action

//  * setThing:     This is the method that's invoked by the slider.
//
//                  if CoalesceP is not set, then it creates a new undo
//                  operation.  One for each twitch of the slider (due to
//                  the slider setting up an undo group, all are grouped
//                  into a single undo operation, but a whole bunch of actions
//                  will get recorded).  This calls changeThing: to actually
//                  change the value and create an undo action
//
//                  if CoalesceP is set, a notification is posted to our private
//                  undoQueue.  When the user stops dragging (during 'idle' time)
//                  the queue will post a notification, which thingChanged:
//                  will pick up.


// change the Thing value in our model to the newValue

- (void) changeThing: (float) newValue
{
    // don't do any work if we're changing it to the previous
    // value.  Otherwise we could end up with no-op actions
    // in the undo stack that would confuse the user (I know it
    // confused me)

    if (newValue != thing) {

	NSLog (@"change thing: %f", newValue);

	NSUndoManager *undoManager;
	undoManager = [window undoManager];

	// create the undo action.  To undo changeThing:, we
	// have to invoke changeThing: but with the previous value
	[[undoManager prepareWithInvocationTarget: self]
	    changeThing: thing];

	// actually change the value in the model.
	// this is not actually recorded by the undo manager.
	thing = newValue;

	// last thing, undo-wise, to do is to set the name
	[undoManager setActionName: @"Change Thing"];
	

	// update the slider, in case we've been invoked from
	// from an undoable action
	[slider setFloatValue: newValue];

	// likewise, update the view.
	[displayField setIntValue: thing];
    }

} // changeThing



// notification handler for @"ThingChanged"

- (void) thingChanged: (NSNotification *) notification
{
    // get the new value out of the slider, which is included as
    // the notification object
    float newThing;
    newThing = [[notification object] floatValue];

    // and actually change the value now, creating an undoable action
    [self changeThing: newThing];

} // thingChanged



// the action method invoked by the slider

- (IBAction) setThing: (id) sender
{
    // snarf the current value from the slider
    float currentValue;
    currentValue = [sender floatValue];

    if (coalesceP) {

	// make a new notification for ThingChanged
	NSNotification *notification;
	notification = [NSNotification notificationWithName: @"ThingChanged"
				       object: sender];


	// post it to our local undo queue.
	// with postingStyle: NSPostWhenIdle, at idle time the
	// notification queue will look at the notifications it has,
	// throw out all the redundant ones (coalesced on name and
	// sender) and then post the remaining notification to the
	// default notification center.  This will ultimately
	// result in a single call to thingChanged: no matter how much
	// the user swings the mouse around

	[undoQueue enqueueNotification: notification
		   postingStyle: NSPostWhenIdle
		   coalesceMask: (NSNotificationCoalescingOnName
				  | NSNotificationCoalescingOnSender)
		   forModes: nil];

	// update our view based on the change, but don't actually
	// change the model.  Here this puts parentheses around the
	// value so that it's obvious while dragging that special
	// tracking-time-only behavior is happening

	NSString *trackingString;
	trackingString = [NSString stringWithFormat: @"(%d)",
				   (int) currentValue];
	[displayField setStringValue: trackingString];

    } else {
	// this is the 'normal' undo case.  Will cause a grouped
	// undo action to be created with every twitch of the mouse.
	// These are all grouped into one action thanks to the slider,
	// but to undo/redo them requires replaying *every* action, which
	// consumes memory and takes time

	[self changeThing: currentValue];
    }

} // setThing



// --------------------------------------------------


// this is architected identically to the *Thing: methods above

- (void) changeColor: (NSColor *) newColor
{
    // don't do any work if we're changing it to the previous
    // value.

    if (![newColor isEqualTo: color]) {

	NSLog (@"change color: %@", newColor);

	// create the undo action.  To undo changeColor:, we have
	// to invoke changeColor: with the previous color
	NSUndoManager *undoManager;
	undoManager = [window undoManager];
	
	[[undoManager prepareWithInvocationTarget: self]
	    changeColor: color];

	// be correct with memory manglement
	[color release];
	color = [newColor copy];
	
	[undoManager setActionName: @"Change Color"];
	
	// update the color well, in case we've been invoked from
	// from an undoable action
	[colorWell setColor: color];

	[displayField setTextColor: color];
	
	// remove any ()'s from the thing value display
	[displayField setIntValue: thing];
    }

} // changeColor



// notification handler for @"ColorChanged"

- (void) colorChanged: (NSNotification *) notification
{
    // get the new value out of the color well
    NSColor *newColor;
    newColor = [[notification object] color];

    // and actually change the value now, creating an undoable action
    [self changeColor: newColor];

} // colorChanged



// the action method invoked by the colorWell

- (IBAction) setColor: (id) sender
{
    // snarf the current value from the color well
    NSColor *currentColor;
    currentColor = [sender color];

    if (coalesceP) {

	// make a new notification for ColorChanged
	NSNotification *notification;
	notification = [NSNotification notificationWithName: @"ColorChanged"
				       object: sender];

	// post it to our local undo queue
	[undoQueue enqueueNotification: notification
		   postingStyle: NSPostWhenIdle
		   coalesceMask: (NSNotificationCoalescingOnName
				  | NSNotificationCoalescingOnSender)
		   forModes: nil];

	// update the view based on the change.  Doesn't actually
	// change the 'model'
	[displayField setTextColor: currentColor];

	// update the tracking string to put the parens around the
	// number so it's obvious the person dragging stuff around that 
	// Something Different is happening
	NSString *trackingString;
	trackingString = [NSString stringWithFormat: @"(%d)",
				   (int) thing];
	[displayField setStringValue: trackingString];
	
    } else {
	[self changeColor: currentColor];
    }

} // setColor



// Other random support stuff


// purge the undo stacks so we can start over from scratch with undoable
// actions

- (IBAction) clearUndo: (id) sender
{
    NSUndoManager *undoManager;
    undoManager = [window undoManager]; // thanks to NSResponder

    [undoManager removeAllActions];

} // clearUndo



// checkbox action, to toggle the coalesce state

- (IBAction) coalesceUndo: (id) sender
{
    coalesceP = ([sender state] == NSOnState);

    NSLog (@"coalesce is %d", coalesceP);

} // coalesceUndo



// --------------------------------------------------

// show all undo-related notifications that come through

- (void) logNotification: (NSNotification *) notification
{
    NSLog (@"%@", [notification  name]);

} // logNotification


@end // AppController


